#include "headers.h"

static int do_send(packlist_t *pl, network_t *net, 
		const char *nick, const char *host,
		pack_t *ptr, int flags, char **ret_str) {
	char list_name[128];
	pack_t *spack	= NULL;
	send_t *ssend	= NULL;
	queue_t *queue;
	int i, ret, pack_no = 0;
	
	if(net == NULL || nick == NULL) {
		if(ret_str != NULL)
			*ret_str = "Invalid Arguments";
		return -1;
	}
	
	LOCK(pl);
	for(i = 0; i < pl->no_packs && spack == NULL; ++i) {
		if(pl->packs[i] == ptr) {
			spack = pack_clone(pl->packs[i]);
			pack_no = i;
		}
	}
	if(spack == NULL) {
		UNLOCK(pl);
		if(ret_str != NULL)
			*ret_str = "Invalid Pack";
		return -1;
	}
	
	xstrncpy(list_name,pl->name,sizeof(list_name));
	
	ssend 		= alloc_send();
	ssend->pack 	= spack;
	ssend->network 	= net;
	ssend->nick 	= xstrdup(nick);
	ssend->host	= host != NULL ? xstrdup(host) : NULL;
	ssend->queue	= pl->queue;
	ssend->retries	= pl->retries;

	queue		= pl->queue;
	
	UNLOCK(pl);

	ret = queue_enqueue(queue,ssend,flags,ret_str);
	if(ret == -1) {
		free_send(ssend);
	} else if(ret >= 0) {
		if(host == NULL) {
			char buffer[128];
			snprintf(buffer,sizeof(buffer),"USERHOST %s",nick);
			irc_send_raw(net,buffer);
		}
		if(ret == 0) {
			LOCK(net);
			LOGP(L_XDC,"SE SEND %s:%d to [%s:%s]",
					list_name,pack_no,net->name,nick);
			UNLOCK(net);
		} else {
			LOCK(net);
			LOGP(L_XDC,"SE SEND %s:%d queued for [%s:%s], postion: %d",
					list_name,pack_no,net->name,nick,ret);
			UNLOCK(net);
		}
		if(ret_str != NULL)
			*ret_str = "Success";
	}

	return ret;
}

static void do_reorder(packlist_t *l, pack_t **packs, int no_packs, int rm) {
	int i,j,found,to_alloc;

	PTRARR_DEL(&packs,&no_packs,NULL);
	
	for(i = 0; i < l->no_packs; ++i) {
		for(j = 0, found = 0; j < no_packs && !found; ++j) {
			if(l->packs[i] == packs[j])
				found = 1;
		}
		if(!found && rm) {
			free_pack(l->packs[i]);
			l->packs[i] = NULL;
		} else if(!found) {
			PTRARR_ADD(&packs,&no_packs,l->packs[i]);
		}
	}

	to_alloc = no_packs;
	for(i = 0; i < no_packs; ++i) {
		for(j = 0, found = 0; j < l->no_packs && !found; ++j) {
			if(packs[i] == l->packs[j])
				found = 1;
		}
		if(!found) {
			packs[i] = NULL;
			to_alloc--;
		}
	}

	xfree(l->packs);
	PTRARR_ALLOC(&(l->packs),&(l->no_packs),to_alloc);

	for(i = 0, j = 0; i < no_packs; ++i) {
		if(packs[i] != NULL)
			l->packs[j++] = packs[i];
	}
}

static void reorder_packs(SV **sp, SV **mark, I32 ax, I32 items,
		packlist_t *l, int rm) {
	pack_t **packs;
	int i, no_packs;;
	
	no_packs = items - 1;
	packs = xalloc(sizeof(pack_t *)*(no_packs+1));
	
	for(i = 0; i < no_packs; ++i) {
		HV *hv = hvref(ST(i+1));
		packs[i] = NULL;
		if(hv != NULL) {
			SV **sv = hv_fetch(hv,"_ptr",4,0);
			if(sv != NULL)
				packs[i] = (pack_t *)SvIV(*sv);
		}
	}
	
	do_reorder(l,packs,no_packs,rm);
	xfree(packs);
}

static pack_t *sv_to_pack(SV *sv) {
	HV *hv;
	SV **psv;

	if((hv = hvref(sv)) == NULL)
		return NULL;

	if((psv = hv_fetch(hv,"_ptr",4,0)) == NULL)
		return NULL;

	return (pack_t *)SvIV(*psv);
}	

MODULE = Arisa::Packlist  PACKAGE = Arisa
PROTOTYPES: ENABLE

Arisa::Packlist
packlist_find(name = NULL)
	char *name
CODE:
	RETVAL = packlist_find(name);
OUTPUT:
	RETVAL

void
packlists()
PREINIT:
	int i;
PPCODE:
	LOCK(global);
	for(i = 0; i < global->no_packlists; ++i) {
		packlist_t *p = global->packlists[i];
		LOCK(p);
		if(p->deleted == 0)
			XPUSHs(sv_2mortal(
				create_ptr_object("Arisa::Packlist",p)));
		UNLOCK(p);
	}
	UNLOCK(global);

Arisa::Packlist
packlist_add(name = NULL, queue)
	char *name
	Arisa::Queue queue
CODE:
	RETVAL = NULL;
	if(name != NULL && queue_valid(queue)) {
		packlist_t *p = alloc_packlist();
		
		p->name = xstrdup(name);
		p->queue = queue;
		
		if(packlist_add(p) == 0) {
			LOCK(p);
			LOCK(p->queue);
			if(p->queue->deleted != 0) {
				UNLOCK(p->queue);
				UNLOCK(p);
				packlist_del(p);
			} else {
				UNLOCK(p->queue);
				UNLOCK(p);
				RETVAL = p;
			}
		} else {
			free_packlist(p);
		}
	}
OUTPUT:
	RETVAL

	
#*******************************
MODULE = Arisa::Packlist  PACKAGE = Arisa::Packlist
#*******************************

SV*
access_add(pl, ...)
	Arisa::Packlist pl
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl)) {
		ACCESS_ADD(pl,pl->access,1);
	}
OUTPUT:
	RETVAL

SV*
access_check(pl, ...)
	Arisa::Packlist pl
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl)) {
		ACCESS_CHECK(pl,pl->access,1);
	}
OUTPUT:
	RETVAL

SV*
access_del(pl, ...)
	Arisa::Packlist pl
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl)) {
		ACCESS_DEL(pl,pl->access,1);
	}
OUTPUT:
	RETVAL
	
void
access_list(pl)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		ACCESS_LIST(pl,pl->access,1);
	}

SV*
access_mode(pl, ...)
	Arisa::Packlist pl
CODE:
	RETVAL = &PL_sv_undef;
	if(packlist_valid(pl)) {
		ACCESS_MODE(pl,pl->access,1);
	}
OUTPUT:
	RETVAL

void
add_pack(pl, ...)
	Arisa::Packlist pl
PREINIT:
	pack_t *ret = NULL;
PPCODE:
	if(packlist_valid(pl)) {
		struct stat sbuf;
		char *file = NULL;
		char *label = NULL;
		int search = 0;
		
		if(items > 2 || (items == 2 && hvref(ST(1)) != NULL)) {
			file = SvPVarg("file");
			label = SvPVarg("label");
			search = SvIVarg("search");
		} else if(items == 2) {
			STRLEN n_a;
			file = SvPV(ST(1),n_a);
		}

		if(file != NULL && search == 0) {
			if(stat(file,&sbuf) == 0)
				pack_add_file(pl,file,label,&sbuf,&ret);
		} else if(file != NULL && search == 1) {
			char buffer[2048], **dirs = NULL;
			int i, no_dirs = 0;
			LOCK(pl);
			dirs = strarr_dup(pl->dirs,pl->no_dirs,&no_dirs);
			UNLOCK(pl);
			for(i = 0; i < no_dirs; ++i) {
				snprintf(buffer,sizeof(buffer),
					"%s%s%s",
					dirs[i],
					dirs[i][strlen(dirs[i])-1] != '/' ? 
						"/" : "",
					file);
				if(stat(buffer,&sbuf) == 0) {
					if(pack_add_file(pl,buffer,label,
							&sbuf,&ret) == 0)
						break;
				}
			}
			if(dirs != NULL)
				xfree(dirs);
		}
	}
	if(ret != NULL) {
		RET_REF(pack_to_hv(ret));
		pack_deref(ret);
	} else {
		RET_REF(NULL);
	}

void
check(pl)
	Arisa::Packlist pl
CODE:
	if(packlist_valid(pl))
		packlist_check(pl);

SV*
del_pack(pl, pack)
	Arisa::Packlist pl
	SV *pack
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl)) {
		pack_t *p = sv_to_pack(pack); 
		if(pack_del(pl,p) == 0)
			RETVAL = &PL_sv_yes;
	}
OUTPUT:
	RETVAL

SV*
delete(pl)
	Arisa::Packlist pl
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl)) {
		if(packlist_del(pl) == 0)
			RETVAL = &PL_sv_yes;
	}
OUTPUT:
	RETVAL
	
void
dirs(pl, ...)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		LOCK(pl);
		LOAD_RET_STRARR(pl->dirs,pl->no_dirs);
		UNLOCK(pl);
	}

void
filters(pl, ...)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		LOCK(pl);
		LOAD_RET_STRARR(pl->filters,pl->no_filters);
		UNLOCK(pl);
	}

void
force_queue(pl, ...)
	Arisa::Packlist pl
PREINIT:
	int	ret_int = -1;
	char 	*ret_str = "Invalid Packlist";
PPCODE:
	if(packlist_valid(pl)) {
		network_t *net	= SvPTRarg("network");
		char *nick	= SvPVarg("nick");
		char *host	= SvPVarg("host");
		pack_t *ptr	= (pack_t *)SvIVarg("_ptr");
		ret_int = do_send(pl,net,nick,host,ptr,
				QUEUE_NO_CLIMIT | QUEUE_NO_ULIMIT,&ret_str);
	}
	XPUSHs(sv_2mortal(newSViv(ret_int)));
	XPUSHs(sv_2mortal(new_pv(ret_str)));
	
void
force_send(pl, ...)
	Arisa::Packlist pl
PREINIT:
	int	ret_int  = -1;
	char 	*ret_str = "Invalid Packlist";
PPCODE:
	if(packlist_valid(pl)) {
		network_t *net	= SvPTRarg("network");
		char *nick	= SvPVarg("nick");
		char *host	= SvPVarg("host");
		pack_t *ptr	= (pack_t *)SvIVarg("_ptr");
		ret_int = do_send(pl,net,nick,host,ptr,SEND_FORCE,&ret_str);
	}
	XPUSHs(sv_2mortal(newSViv(ret_int)));
	XPUSHs(sv_2mortal(new_pv(ret_str)));

SV*
name(pl, new_name = NULL)
	Arisa::Packlist pl
	char *new_name
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl)) {
		if(new_name != NULL) {
			if(packlist_rename(pl,new_name) == 0)
				RETVAL = &PL_sv_yes;
		} else {
			LOCK(pl);
			RETVAL = new_pv(pl->name);
			UNLOCK(pl);
		}
	}
OUTPUT:
	RETVAL

SV*
ignore_add(pl, ...)
	Arisa::Packlist pl
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl)) {
		IGNORE_LIST_ADD(pl,pl->ignore_list);
	}
OUTPUT:
	RETVAL

SV*
ignore_del(pl, ...)
	Arisa::Packlist pl
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl)) {
		IGNORE_LIST_DEL(pl,pl->ignore_list);	
	}
OUTPUT:
	RETVAL

void
ignore_list(pl, ...)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		IGNORE_LIST_LIST(pl,pl->ignore_list);
	}

SV*
ignore_move(pl, src, dst)
	Arisa::Packlist pl
	int src
	int dst
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl)) {
		IGNORE_LIST_MOVE(pl,pl->ignore_list,src,dst);
	}
OUTPUT:
	RETVAL

void
ignore_settings(pl, ...)
	Arisa::Packlist pl
PREINIT:
	HV *RETVAL = NULL;
PPCODE:
	if(packlist_valid(pl)) {
		IGNORE_LIST_SETTINGS(pl,pl->ignore_list);
	}
	RET_REF(RETVAL);

SV*
ignore_test(pl, host)
	Arisa::Packlist pl
	char *host
CODE:
	RETVAL = &PL_sv_no;
	if(packlist_valid(pl) && host != NULL) {
		LOCK(pl);
		if(ignore_test(pl->ignore_list,host))
			RETVAL = &PL_sv_yes;
		UNLOCK(pl);
	}
OUTPUT:
	RETVAL
	
void
ignore_wipe(pl)
	Arisa::Packlist pl
CODE:
	if(packlist_valid(pl)) {
		IGNORE_LIST_WIPE(pl,pl->ignore_list);
	}

void
info(pl)
	Arisa::Packlist pl
PREINIT:
	HV *hv = NULL;
PPCODE:
	if(packlist_valid(pl)) {
		hv = newHV();
		LOCK(pl);
		hv_store(hv,"last_scan",9,newSVuv(pl->last_scan),0);
		UNLOCK(pl);
	}
	RET_REF(hv);

SV*
label(pl, pack, label = NULL)
	Arisa::Packlist pl
	SV *pack
	char *label
PREINIT:
	pack_t *p;
CODE:
	p = sv_to_pack(pack);
	RETVAL = &PL_sv_no;
	if(pack_valid(pl,p)) {
		WLOCK(p);
		if(p->label != NULL)
			xfree(p->label);
		if(label == NULL) {
			char buffer[2048];
			pathtofn(p->file,buffer,sizeof(buffer));
			p->label = xstrdup(buffer);
		} else {
			p->label = xstrdup(label);
		}
		WUNLOCK(p);
		pack_deref(p);
		RETVAL = &PL_sv_yes;
	}
OUTPUT:
	RETVAL

void
nolist(pl, ...)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		LOCK(pl);
		if(items > 1) {
			pl->nolist = SvUV(ST(1));
			if(pl->nolist_reason != NULL) {
				xfree(pl->nolist_reason);
				pl->nolist_reason = NULL;
			}
			if(items > 2) {
				STRLEN n_a;
				char *reason = SvPV(ST(2),n_a);
				if(reason != NULL)
					pl->nolist_reason = xstrdup(reason);
			}
		}
		if(pl->nolist != 0) {
			XPUSHs(sv_2mortal(newSVuv(pl->nolist)));
			XPUSHs(sv_2mortal(new_pv(pl->nolist_reason)));
		}
		UNLOCK(pl);
	}

void
nosend(pl, ...)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		LOCK(pl);
		if(items > 1) {
			pl->nosend = SvUV(ST(1));
			if(pl->nosend_reason != NULL) {
				xfree(pl->nosend_reason);
				pl->nosend_reason = NULL;
			}
			if(items > 2) {
				STRLEN n_a;
				char *reason = SvPV(ST(2),n_a);
				if(reason != NULL)
					pl->nosend_reason = xstrdup(reason);
			}
		}
		if(pl->nosend != 0) {
			XPUSHs(sv_2mortal(newSVuv(pl->nosend)));
			XPUSHs(sv_2mortal(new_pv(pl->nosend_reason)));
		}
		UNLOCK(pl);
	}

void
packs(pl)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		int i;

		LOCK(pl);

		for(i = 0; i < pl->no_packs; ++i) {
			RLOCK(pl->packs[i]);
			XPUSHs(sv_2mortal(newRV_noinc(
				(SV*)pack_to_hv(pl->packs[i]))));
			RUNLOCK(pl->packs[i]);
		}

		UNLOCK(pl);
	}


void
plist(pl, ...)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		char 		list_name[128];
		pformat_t	pformat_type	= PFORMAT_FULL;
		listtype_t	list_type	= LIST_NORMAL;;
		colourtype_t	colour_type	= COLOUR_IRC;
		responsetype_t	response_type;

		network_t *net  	= SvPTRarg("network");
		channel_t *chan		= NULL;
		SV *chan_sv		= SVarg("channel");
		char *nick		= SvPVarg("nick");
		char *host		= SvPVarg("host");
		char *hostmask		= SvPVarg("hostmask");
		char *type		= SvPVarg("type");
		char *pformat		= SvPVarg("pformat");
		char *colour		= SvPVarg("colour");
		
		LOCK(pl);
		xstrncpy(list_name,pl->name,sizeof(list_name));
		list_type	= pl->type;
		response_type	= pl->response;
		UNLOCK(pl);
		
		if(type != NULL) {
			if((list_type = str_to_listtype(type)) == -1)
				list_type = LIST_NORMAL;
		}
		if(pformat != NULL) {
			if((pformat_type = str_to_pformat(pformat)) == -1)
				pformat_type = PFORMAT_FULL;
		}
		if(colour != NULL) {
			if((colour_type = str_to_colourtype(colour)) == -1)
				colour_type = COLOUR_IRC;
		}

		if(!irc_network_valid(net))
			net = NULL;
		
		if(chan_sv != NULL) {
			ptr2_t ptr;
			fetch_object_ptr2(chan_sv,&ptr);
			if(irc_channel_valid(ptr.minor,ptr.major)) {
				net = ptr.minor;
				chan = ptr.major;
			}
		}

		if(list_type != LIST_CHAT) {
			char *msg;
			pqueue_t listing;
			
			pqueue_init(&listing,0);
			
			if(pformat == NULL)
				packlist_xdl_generate(pl,&listing,
					list_type,colour_type,net,chan,"","");
			else
				packlist_plist_generate(pl,&listing,
					pformat_type,colour_type,net,chan,"","");
			
			if(nick != NULL && net != NULL) {
				while((msg = pqueue_pop_front(&listing)) != NULL) {
					if(response_type == RESPONSE_NOTICE)
						irc_send_notice(net,nick,msg);
					else if(response_type == RESPONSE_PRIVMSG)
						irc_send_msg(net,nick,msg);
					XPUSHs(sv_2mortal(new_pv(msg)));
					xfree(msg);
				}
				LOCK(net);
				LOGP(L_XDC,"XDCC LIST of %s sent to [%s:%s] (SE request)",
						list_name,net->name,nick);
				UNLOCK(net);
			} else if(chan != NULL && net != NULL) {
				char cname[64];
				
				LOCK(chan);
				xstrncpy(cname,chan->name,sizeof(cname));
				UNLOCK(chan);
				
				while((msg = pqueue_pop_front(&listing)) != NULL) {
					irc_send_msg(net,cname,msg);
					XPUSHs(sv_2mortal(new_pv(msg)));
					xfree(msg);
				}
				LOCK(net);
				LOGP(L_XDC,"XDCC LIST of %s sent to [%s:%s] (SE request)",
						list_name,net->name,cname);
				UNLOCK(net);
			} else {
				while((msg = pqueue_pop_front(&listing)) != NULL) {
					XPUSHs(sv_2mortal(new_pv(msg)));
					xfree(msg);
				}
			}
		} else {
			if(chat_list_open(net,host,hostmask,nick,pl) == 0)
				XPUSHs(&PL_sv_yes);
			else
				XPUSHs(&PL_sv_no);
		}
	}		
		
	
Arisa::Queue
queue(pl, q = NULL)
	Arisa::Packlist pl
	Arisa::Queue q
CODE:
	RETVAL = NULL;
	if(packlist_valid(pl)) {
		if(q != NULL && queue_valid(q)) {
			queue_t *tmp;
			LOCK(pl);
			tmp = pl->queue;
			pl->queue = q;
			LOCK(pl->queue);
			if(pl->queue->deleted != 0) {
				UNLOCK(pl->queue);
				pl->queue = tmp;
			} else {
				UNLOCK(pl->queue);
			}
			RETVAL = pl->queue;
			UNLOCK(pl);
		} else {
			LOCK(pl);
			RETVAL = pl->queue;
			UNLOCK(pl);
		}
	}
OUTPUT:
	RETVAL

SV*
rehash(pl, pack)
	Arisa::Packlist pl
	SV *pack
PREINIT:
	pack_t *p;
CODE:
	p = sv_to_pack(pack);
	RETVAL = &PL_sv_no;
	if(pack_valid(pl,p)) {
		WLOCK(p);
		if(p->crc != NULL) {
			free(p->crc);
			p->crc = NULL;
		}
		if(p->md5 != NULL) {
			xfree(p->md5);
			p->md5 = NULL;
		}
		p->hashed = 0;
		WUNLOCK(p);
		hash_request(p);
		pack_deref(p);
		RETVAL = &PL_sv_yes;
	}
OUTPUT:
	RETVAL

void
reorder(pl, ...)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		int i;

		LOCK(pl);

		reorder_packs(sp,mark,ax,items,pl,0);

		for(i = 0; i < pl->no_packs; ++i) {
			RLOCK(pl->packs[i]);
			XPUSHs(sv_2mortal(newRV_noinc(
				(SV*)pack_to_hv(pl->packs[i]))));
			RUNLOCK(pl->packs[i]);
		}

		UNLOCK(pl);
	}

void
reorder_rm(pl, ...)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		int i;

		LOCK(pl);

		reorder_packs(sp,mark,ax,items,pl,1);

		for(i = 0; i < pl->no_packs; ++i) {
			RLOCK(pl->packs[i]);
			XPUSHs(sv_2mortal(newRV_noinc(
				(SV*)pack_to_hv(pl->packs[i]))));
			RUNLOCK(pl->packs[i]);
		}

		UNLOCK(pl);
	}

void
scan(pl, ...)
	Arisa::Packlist pl
CODE:
	if(packlist_valid(pl)) {
		int depth = 0;
		if(items > 1)
			depth = SvIV(ST(1));
		packlist_scan(pl,depth);
	}

void
send(pl, ...)
	Arisa::Packlist pl
PREINIT:
	int	ret_int  = -1;
	char 	*ret_str = "Invalid Packlist";
PPCODE:
	if(packlist_valid(pl)) {
		network_t *net	= SvPTRarg("network");
		char *nick	= SvPVarg("nick");
		char *host	= SvPVarg("host");
		pack_t *ptr	= (pack_t *)SvIVarg("_ptr");
		ret_int = do_send(pl,net,nick,host,ptr,0,&ret_str);
	}
	XPUSHs(sv_2mortal(newSViv(ret_int)));
	XPUSHs(sv_2mortal(new_pv(ret_str)));
	
void
settings(pl, ...)
	Arisa::Packlist pl
PREINIT:
	HV *hv = NULL;
PPCODE:
	if(packlist_valid(pl)) {
		hv = newHV();
		perl_apply_settings(packlist_apply_setting,pl);
		perl_read_settings(packlist_read_settings,pl,hv);
	}
	RET_REF(hv);

void
taglines(pl, ...)
	Arisa::Packlist pl
PPCODE:
	if(packlist_valid(pl)) {
		LOCK(pl);
		LOAD_RET_STRARR(pl->taglines,pl->no_taglines);
		UNLOCK(pl);
	}
